#!/usr/bin/perl -w -T
#
# Library of procedures for use 
# by the EduRoam cgi-bin pages
#

package eduroam;

BEGIN { }

use strict;
use Net::SNMP;
use DBI;

##########################################
# AAI
our $ldap_root_password = "";
our %ldap_userdb;
our %radius_userdb;
# These ones are derived from other data
our $dc_domain;
our $root_dn;
our $root_pw;
our $domain;
our $user;
our $user_password;

# ACC
our $radiusd_pw = "";
our $dhcpd_pw = "";
our $apd_pw = "";
our $rad_clean_pw = "";

# SYS_ACL
our @sys_webacls = ();

our $dhcp_enable = 1;

# DHCP_DOMAIN_NAME
our $dhcp_domain_name = "";

# DHCP_DNS
our @dhcp_dns = ();

# DHCP_POOL
# dhcp is a complex record with the following attributes (in an array of hashes):
# 1. network - name of shared network
# 2. cidr    - ip address in CIDR format
# 3. gw      - default route for CIDR
# 4. range   - string with ranges of IPs (TODO: currently only 1 range per subnet)
our @dhcp_pool = ();

# CERT_SERVER
our $cert_san; # Subject Alternative Name to be put in a certificate request
our $httpd_use_ssl = 0;

# AAI_REALM
our $aai_realm = "";

# AAI_IDP
our $idp_admin_name = "";
our $idp_admin_email = "";
our $idp_admin_password = "";
our $idp_host = "";
our $idp_policy_url = "";
our $idp_language = "";
our @idp_all_languages = ('en', 'no', 'nn', 'se', 'fi', 'da', 'sv', 'de', 'es', 'fr', 'nl', 'lb', 'hr', 'hu', 'pl', 'sl', 'pt', 'pt-BR');
our $idp_org_name = "";
our $idp_description = "";
our $idp_meta_email = "";
our $idp_use_ds = '0';
our $idp_ds_url = "https://ds.aai.arnes.si/simplesaml/saml2/sp/idpdisco.php";
our $idp_fed_url = "http://ds.aai.arnes.si/federation/metadata/aai.arnes.si.xml";

our $idp_secretsalt = "";
our $idp_uri_date = "";
our $idp_uri_idp = "";
our $idp_uri_sp = "";

# AAI_RAD_PEERS
our $primary_radius_ip = "";
our $primary_radius_sec = "";
our $secondary_radius_ip = "";
our $secondary_radius_sec = "";

# AP
# work variables
our $sql_ap_dbh;
our $sql_ap_query;
our $sql_ap_update;
our $sql_ap_insert;
our $sql_ap_delete;
our %sql_ap_aps;

##########################################
# Configs
our %config_files;
our %component_cfgs;
our %component_description;
our @component_build_order;
our %components;
our $config_path;
our $flag_file;

##########################################
# variable initialisation
sub BEGIN {
	# config/commit
	%eduroam::config_files = (
		"sys_db"                => ["radius", "radius clients", "eduroam monitor"],
		"cert_server"           => ["httpd"],
		"dhcp_cidr"             => ["dhcp"],
		"dhcp_domain_name"      => ["dhcp"],
		"dhcp_dns"              => ["dhcp"],
		"dhcp_pool"             => ["dhcp"],
		"aai_realm"             => ["radius", "radius proxy", "radius users", "ldap", "ldap directory", "simplesamlphp"],
		"aai_rad_peers"         => ["radius proxy", "radius clients"],
		"aai_rad_users"         => ["radius users"],
		"aai_ldap_password"     => ["ldap", "ldap directory", "simplesamlphp", "radius"],
		"aai_ldap_users"        => ["ldap directory"],
		"aai_idp"               => ["simplesamlphp"],
		"ap"                    => ["radius clients"],
        );
	%eduroam::component_description = (
		"dhcp"                  => "DHCP server configuration",
		"ldap"                  => "LDAP server configuration",
		"ldap directory"        => "Configuration of data in LDAP directory",
		"simplesamlphp"         => "simpleSAMLphp configuration as an Identity Provider",
		"radius clients"        => "RADIUS client devices and peers",
		"radius proxy"          => "RADIUS server request proxying",
		"radius users"          => "Staticly configured users in RADIUS settings",
		"radius"                => "RADIUS server general configuration",
		"eduroam monitor"       => "Complement accounting with IP information and close old accounting records",
		"httpd"                 => "Configure apache to use SSL",
        );
	@eduroam::component_build_order = (
		"dhcp",
		"ldap",
		"ldap directory",
		"simplesamlphp",
		"radius clients",
		"radius proxy",
		"radius users",
		"radius",
		"eduroam monitor",
		"httpd",
	);
	%eduroam::component_cfgs = ();
	%eduroam::components = ();
	$eduroam::config_path = "/var/eduroam/etc/";
	$eduroam::flag_file = "commit.flag";

	# AAI
	$eduroam::aai_realm = "";
	$eduroam::ldap_root_password = "";
	$eduroam::primary_radius_ip = "";
	$eduroam::primary_radius_sec = "";
	$eduroam::secondary_radius_ip = "";
	$eduroam::secondary_radius_sec = "";
}

##########################################
#########################################
# Input data validators
# 1st parameter = range
# 2nd parameter = cidr
# return 0 if OK
#########################################
sub validate_range {
	my $range = $_[0];
	my $minip;
	my $maxip;
	my $ip;
	my $result = 1;
	my @ranges;
	my $tmp;
	my $ip1;
	my $ip2;

	return 1 unless defined $_[0];
	return 1 unless defined $_[1];
	return 1 if length $_[0] <= 0;
	return 1 if length $_[1] <= 0;
	($minip, $maxip) = &eduroam::IpCidr2MinMaxIP($_[1]);
	# minip and maxip are already out of network/broadcast addresses...
	$minip = &eduroam::IpStr2Int($minip);
	$maxip = &eduroam::IpStr2Int($maxip);

	# range is in format: ip1-ip2, ip3-ip4, ...
	# split on comma
	@ranges = split(/,/,$range);
	foreach $tmp (@ranges) {
		return 1 unless $tmp =~ m/^\s*(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s*\-\s*(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s*$/;
		$ip1 = &eduroam::IpStr2Int($1);
		$ip2 = &eduroam::IpStr2Int($2);
		return 1 if $ip2 < $ip1;
		return 1 if $ip1 < $minip;
		return 1 if $ip1 > $maxip;
		return 1 if $ip2 < $minip;
		return 1 if $ip2 > $maxip;
	}
	return 0;
}
#########################################
sub validate_hostname {
	my $hostname = $_[0];
	return 1 unless ($hostname =~ m/^[\w\.\-]*$/g);
	return 0;
}
#########################################
sub validate_description {
	my $description = $_[0];
	return 1 if (not $description =~ m/^[\w]*$/g);
	return 0;
}
#########################################
sub validate_vlan {
	my $vlan = $_[0];
	return 1 if (not $vlan =~ m/^[\d]+$/g);
	$vlan = 0 + $vlan; # make number
	return 1 if ($vlan < 1 or $vlan > 2048);
	return 0;
}
#########################################
sub validate_username {
	my $username = $_[0];
	return 1 if (not $username =~ m/^[\w]*$/g);
	return 0;
}
#########################################
sub validate_password {
	my $pass1 = $_[0];
	my $pass2 = $_[1];
	if (defined($pass2)) {
		return 2 if ($pass1 ne $pass2)
	}
	return 1 if (not $pass1 =~ m/^[\w]*$/);
	return 0;
}
#########################################
sub validate_cidr {
	my $tmp = $_[0];
	my $ip = undef;
	my $cidr = undef;

	return 1 if not defined ($tmp);
	if ($tmp =~ m/^(.*)\/(\d{1,2})$/) {
		$ip = $1;
		$cidr = 0+$2;
	} else {
		$ip = $tmp;
		$cidr = 32;
	}
        return 1 unless ($cidr>=0 && $cidr<=32);
	return &validate_ip($ip);
}
#########################################
sub validate_ip {
	my $tmp = $_[0];
	my $ip1;
	my $ip2;
	my $ip3;
	my $ip4;
	
	return 1 if not defined ($tmp);
	return 1 unless ($tmp =~ m/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/);
	$ip1 = 0 + $1;
	$ip2 = 0 + $2;
	$ip3 = 0 + $3;
	$ip4 = 0 + $4;
	return 1 unless (0<=$ip1 && $ip1<=255);
	return 1 unless (0<=$ip2 && $ip2<=255);
	return 1 unless (0<=$ip3 && $ip3<=255);
	return 1 unless (0<=$ip4 && $ip4<=255);
	return 0;
}
#########################################
#########################################
# MySQL data access
#########################################
#########################################
# Establish connection for APs and prepare
# queries. Ap config file muste be loaded.
# return query templates
sub sql_ap_connect {
	my $dsn = 'DBI:mysql:radius:localhost';
	$eduroam::sql_ap_dbh = DBI->connect($dsn, 'root', $eduroam::sql_root_pw);
	return undef unless (defined $eduroam::sql_ap_dbh);
	$eduroam::sql_ap_query = $eduroam::sql_ap_dbh->prepare(q{SELECT `IP address`,`snmp secret`,
	`radius secret`,`root username`,`root password` FROM access_points});
	$eduroam::sql_ap_update = $eduroam::sql_ap_dbh->prepare(q{UPDATE access_points SET
	`snmp secret` = ?, `radius secret` = ?,
	`root username` = ?, `root password` = ?
	WHERE `IP address` = ?});
	$eduroam::sql_ap_insert = $eduroam::sql_ap_dbh->prepare(q{INSERT access_points SET
	`snmp secret` = ?, `radius secret` = ?,
	`root username` = ?, `root password` = ?,
	`IP address` = ?});
	$eduroam::sql_ap_delete = $eduroam::sql_ap_dbh->prepare(q{DELETE FROM access_points WHERE `IP address` = ?});
	return $eduroam::sql_ap_dbh;
}
sub sql_ap_loadall {
	my $result;

	$eduroam::sql_ap_query->execute() || die $eduroam::sql_ap_dbh->errstr;
        %eduroam::sql_ap_aps = ();
        while ($result = $eduroam::sql_ap_query->fetchrow_arrayref) {
                $eduroam::sql_ap_aps{$result->[0]} = {	'snmp' => $result->[1],
						'radius' => $result->[2],
						'username' => $result->[3],
						'password' => $result->[4],
		};
	}
	$eduroam::sql_ap_query->finish;
	return;
}
sub sql_ap_disconnect {
	$eduroam::sql_ap_dbh->disconnect;
}
#########################################
#########################################
# Handling of configuration files/modules
# We only put up a flag that tells us when the script was "committed"
sub set_flagfile {
        open(FH, ">$eduroam::config_path"."$eduroam::flag_file");
        close(FH);
}
##########################################
sub get_component_cfgs {
	my $component = $_[0];
	my %tmp_component;
	my @cfgs = ();
	my $tmp;
	my $tmp2;
	my $tmp3;
        
	foreach $tmp (keys %eduroam::config_files) {
		foreach $tmp2 (@{$eduroam::config_files{$tmp}}) {
			# build dependency information:
			# for every module -> which config file must be loaded ?!
			# traverse %config_files and build inverse of it.
			$tmp3 = $eduroam::component_cfgs{$tmp2}; # component has a list of matching config files
			$tmp3 = {} unless defined $tmp3;
			$tmp3->{$tmp} = ""; # insert this config file into the hash of config files belonging to a component
			$eduroam::component_cfgs{$tmp2} = $tmp3;
                }
        }
	
	%tmp_component = %{$eduroam::component_cfgs{$component}};
	foreach $tmp (keys %tmp_component) {
		push @cfgs, $tmp;
	}

	return @cfgs;
}
##########################################
# Check which of the components need to be updated
sub check_components_state {
        # Read from %config_files a list of components and
        # fill them into the %components. Add a state as Pending
        # if there was a change in one of the configuration files.

	my $tmp;
	my $tmp2;
	my $tmp3;

        my @stat = stat "$eduroam::config_path" . "$eduroam::flag_file";
        my $flag_mtime = -1;
        my $cfg_mtime = -2;
        $flag_mtime = $stat[10] if defined $stat[10];

        foreach $tmp (keys %eduroam::config_files) {
                @stat = stat "$eduroam::config_path" . "$tmp";
                $cfg_mtime = $stat[10] if defined $stat[10];

                # Generate sorted components list from the config_files hash
                foreach $tmp2 (@{$eduroam::config_files{$tmp}}) {
                        # Default is OK, we reset it to Pending if times don't match
                        # Mind that multiple config files set these flags.
                        # OK must not overwrite Pending. Pending must overwrite OK.
                        $eduroam::components{$tmp2} = "OK" unless defined $eduroam::components{$tmp2};
                        $eduroam::components{$tmp2} = "Pending" unless ($flag_mtime >= $cfg_mtime);
			# build dependency information:
			# for every module -> which config file must be loaded ?!
			# traverse %config_files and build inverse of it.
			$tmp3 = $eduroam::component_cfgs{$tmp2}; # component has a list of matching config files
			$tmp3 = {} unless defined $tmp3;
			$tmp3->{$tmp} = ""; # insert this config file into the hash of config files belonging to a component
			$eduroam::component_cfgs{$tmp2} = $tmp3;
                }
        }
}
#########################################
# check_components_state must be run
# before this procedure.
sub commit_changes {
	my $tmp;
	my $tmp2;
	my $component_script;
	my $rebuild_all = 0;
	$rebuild_all = $_[0] if defined $_[0];
	
	# for every indicated component run a re-build
	foreach $tmp (@eduroam::component_build_order) {
		next unless $rebuild_all or $eduroam::components{$tmp} eq "Pending";
		# execute a module update function
		$component_script = $tmp;
		$component_script =~ s/ /_/g;
		print "<hr><b>Executing script $component_script ($tmp)</b><br>";
		&eduroam::system_or_htmldie("/var/eduroam/helpers/eduroam_helper $component_script");
	}
	&eduroam::set_flagfile;
}
#########################################
#########################################
# Substitutes variable in a line from a template file
# Returns an array of processed (values substituted) lines.
sub process_template_line {
        my @output = ();
        my $line = $_[0];
        my $tmp;

        # Add the line to the output only if there are no variables
	if (not $line =~ m/<%\w+%>/) {
#        if (not $line =~ m/(^\$[^\$]|[^\$]\$[^\$]|[^\$]\$$)/) {
# 		$line =~ s/\${2}/\$/g; # handle escape sequence $$->$
		push @output, $line;
		return @output;
	}

	##########################################################################
	# SYS_ACL componentes
        
	# <%sys_webacl%>
        # -- all ACLs for access to web interface 
        if ($line =~ m/<%sys_webacl%>/) {
		my $acl;
		foreach $acl (@eduroam::sys_webacls) {
			$tmp = $line;
			$tmp =~ s/<%sys_webacl%>/$acl/;
			push @output, &process_template_line($tmp);
		}
                return @output;
        }

	##########################################################################
	# CERT_SERVER componentes
        
	# <%cert_san%>
	# Subject Alternative Name for openssl config
        if ($line =~ m/<%cert_san%>/) {
                $tmp = $line;
                $tmp =~ s/<%cert_san%>/$eduroam::cert_san/;
                push @output, &process_template_line($tmp);
                return @output;
        }

	##########################################################################
	# AAI componentes
		
	# <%radius_server%>
        # -- ACLs for new connections from other radius servers
        if ($line =~ m/<%radius_server%>/) {
		if(length($eduroam::primary_radius_ip) > 0) {
			$tmp = $line;
			$tmp =~ s/<%radius_server%>/$eduroam::primary_radius_ip/;
			push @output, &process_template_line($tmp);
		}
		if(length($eduroam::secondary_radius_ip) > 0) {
			$tmp = $line;
			$tmp =~ s/<%radius_server%>/$eduroam::secondary_radius_ip/;
			push @output, &process_template_line($tmp);
		}
                return @output;
        }

        # <%dc_domain%>
        if ($line =~ m/<%dc_domain%>/) {
                $tmp = $line;
                $tmp =~ s/<%dc_domain%>/$eduroam::dc_domain/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%root_dn%>
        if ($line =~ m/<%root_dn%>/) {
                $tmp = $line;
                $tmp =~ s/<%root_dn%>/$eduroam::root_dn/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%root_pw%>
        if ($line =~ m/<%root_pw%>/) {
                $tmp = $line;
                $tmp =~ s/<%root_pw%>/$eduroam::root_pw/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%domain%>
        if ($line =~ m/<%domain%>/) {
                $tmp = $line;
                $tmp =~ s/<%domain%>/$eduroam::domain/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%full_domain%>
        if ($line =~ m/<%full_domain%>/) {
                $tmp = $line;
                $tmp =~ s/<%full_domain%>/$eduroam::aai_realm/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%user_password%>
        if ($line =~ m/<%user_password%>/) {
                $tmp = $line;
                $tmp =~ s/<%user_password%>/$eduroam::user_password/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%user%>
        if ($line =~ m/<%user%>/) {
                $tmp = $line;
                $tmp =~ s/<%user%>/$eduroam::user/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        # <%rad_user_user%> in <%rad_user_pass%>
        # -- matching of radius username and password 
        if ($line =~ m/<%rad_user_name%>/ and $line =~ m/<%rad_user_pass%>/) {
                my $rad_user;
                foreach $rad_user (keys %eduroam::radius_userdb) {
                        $tmp = $line;
                        $tmp =~ s/<%rad_user_name%>/$rad_user/;
                        $tmp =~ s/<%rad_user_pass%>/$eduroam::radius_userdb{$rad_user}/;
                        push @output, &process_template_line($tmp);
                }
        }

	# <%idp_admin_name%>
	if ($line =~ m/<%idp_admin_name%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_admin_name%>/$eduroam::idp_admin_name/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_admin_email%>
	if ($line =~ m/<%idp_admin_email%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_admin_email%>/$eduroam::idp_admin_email/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_admin_password%>
	if ($line =~ m/<%idp_admin_password%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_admin_password%>/$eduroam::idp_admin_password/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_host%>
	if ($line =~ m/<%idp_host%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_host%>/$eduroam::idp_host/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_policy_url%>
	if ($line =~ m/<%idp_policy_url%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_policy_url%>/$eduroam::idp_policy_url/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_language%>
	if ($line =~ m/<%idp_language%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_language%>/$eduroam::idp_language/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_secretsalt%>
	if ($line =~ m/<%idp_secretsalt%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_secretsalt%>/$eduroam::idp_secretsalt/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_uri_date%>
	if ($line =~ m/<%idp_uri_date%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_uri_date%>/$eduroam::idp_uri_date/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_uri_idp%>
	if ($line =~ m/<%idp_uri_idp%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_uri_idp%>/$eduroam::idp_uri_idp/;
		push @output, &process_template_line($tmp);
		return @output;
	}
	
	# <%idp_uri_sp%>
	if ($line =~ m/<%idp_uri_sp%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_uri_sp%>/$eduroam::idp_uri_sp/;
		push @output, &process_template_line($tmp);
		return @output;
	}

	# <%idp_org_name%>
	if ($line =~ m/<%idp_org_name%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_org_name%>/$eduroam::idp_org_name/;
		push @output, &process_template_line($tmp);
		return @output;
	}

	# <%idp_description%>
	if ($line =~ m/<%idp_description%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_description%>/$eduroam::idp_description/;
		push @output, &process_template_line($tmp);
		return @output;
	}

	# <%idp_meta_email%>
	if ($line =~ m/<%idp_meta_email%>/) {
		$tmp = $line;
		$tmp =~ s/<%idp_meta_email%>/$eduroam::idp_meta_email/;
		push @output, &process_template_line($tmp);
		return @output;
	}

	# <%idp_uri_sp%>
	if ($line =~ m/<%idp_ds_url%>/) {
		$tmp = $line;
		if ($eduroam::idp_use_ds) {
			$tmp =~ s/<%idp_ds_url%>/\'$eduroam::idp_ds_url\'/;
		} else {
			$tmp =~ s/<%idp_ds_url%>/NULL/;
		}
		push @output, &process_template_line($tmp);
		return @output;
	}

	# <%idp_default%>
	if ($line =~ m/<%idp_default%>/) {
		$tmp = $line;
		if ($eduroam::idp_use_ds) {
			$tmp =~ s/<%idp_default%>/NULL/;
		} else {
			$tmp =~ s/<%idp_default%>/\'$eduroam::idp_uri_idp\'/;
		}
		push @output, &process_template_line($tmp);
		return @output;
	}

	##########################################################################

	##########################################################################
	# ACC componentes

	# <%radiusd_password%>
        if ($line =~ m/<%radiusd_password%>/) {
                $tmp = $line;
                $tmp =~ s/<%radiusd_password%>/$eduroam::radiusd_pw/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        if ($line =~ m/<%dhcpd_password%>/) {
                $tmp = $line;
                $tmp =~ s/<%dhcpd_password%>/$eduroam::dhcpd_pw/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        if ($line =~ m/<%apd_password%>/) {
                $tmp = $line;
                $tmp =~ s/<%apd_password%>/$eduroam::apd_pw/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        if ($line =~ m/<%denied_password%>/) {
                $tmp = $line;
                $tmp =~ s/<%denied_password%>/$eduroam::denied_pw/;
                push @output, &process_template_line($tmp);
                return @output;
        }

        if ($line =~ m/<%rad_clean_password%>/) {
                $tmp = $line;
                $tmp =~ s/<%rad_clean_password%>/$eduroam::rad_clean_pw/;
                push @output, &process_template_line($tmp);
                return @output;
        }
	##########################################################################

	##########################################################################
	# AP componentes
        if ($line =~ m/<%ap%>/) {
		my $ap;
		foreach $ap (keys %eduroam::aps) {
                	$tmp = $line;
	                $tmp =~ s/<%ap%>/$ap/;
	                push @output, &process_template_line($tmp);
		}
                return @output;
        }
	##########################################################################

        return @output;
}
##########################################
sub read_htaccess_webacls {
	my $htaccess = '/var/eduroam/cgi-bin/.htaccess';
	my @output;
	my $tmp;
	
	@eduroam::sys_webacls = ();
	push @eduroam::sys_webacls, "127.0.0.1"; # always allow localhost
	return if not open FH, "<$htaccess";
	while (<FH>) {
		next if not $_ =~ m/Allow from (.*)/i;
		$tmp = $1;
		next if $tmp =~ m/127.0.0.1/;
		push @eduroam::sys_webacls, $tmp;
	}
	close(FH);
}
##########################################
sub write_htaccess {
	my $htaccess_template = '/var/eduroam/system_templates/htaccess.eduroam';
	my $htaccess = '/var/eduroam/cgi-bin/.htaccess';
	my $location = $_[0];
	my @output;

	&eduroam::html_perror2("Can't open source file", $location) if not open FH, "<$htaccess_template";
	while(<FH>) {
		push @output, &eduroam::process_template_line($_);
	}
	&eduroam::html_perror2("Can't open destination file", $location) if not open FH, ">$htaccess";
	print FH @output;
	close(FH);
}
##########################################
sub process_aai {
	$eduroam::dc_domain = $eduroam::aai_realm;
	$eduroam::dc_domain =~ s/\./,dc=/g;
	$eduroam::dc_domain =~ s/^/dc=/g;
	$eduroam::aai_realm =~ m/^(\w+)\./;
	$eduroam::domain = $1;
	$eduroam::root_dn = "cn=root,$eduroam::dc_domain";
	$eduroam::root_pw = "$eduroam::ldap_root_password";

	# simpleSAMLphp entity ID URIs
	$eduroam::idp_uri_idp = "https://$idp_host/idp/$eduroam::idp_uri_date";
	$eduroam::idp_uri_sp  = "https://$idp_host/sp2/$eduroam::idp_uri_date";
}

##########################################
# Load simpleSAMLphp metadata from federation
# return 0 on success, 1 on fail
#
sub write_fed_metadata {
	my $url = shift;
	my @idp_lines;
	my @sp_lines;
	my $idp_f;
	my $sp_f;
	my $ssp_root = "/var/simplesamlphp";
	my $metarefresh = "$ssp_root/modules/metarefresh/bin/metarefresh.php";
	my $idp_meta_conf = "$ssp_root/metadata/saml20-idp-remote.php";
	my $sp_meta_conf = "$ssp_root/metadata/saml20-sp-remote.php";
	my $line;

	$url =~ m/(.*)/;
	$url = $1;

	open (FH, "/usr/bin/php $metarefresh -s $url |") or return 1;
	# sort metadata
	while(<FH>) {
		$line = $_;
		$idp_f = 0 if ($line =~ m/End of data.*saml20-idp-remote\.php/);
		$sp_f  = 0 if ($line =~ m/End of data.*saml20-sp-remote\.php/);

		push @idp_lines, $line if ($idp_f);
		push @sp_lines,  $line if ($sp_f);

		$idp_f = 1 if ($line =~ m/The following.*saml20-idp-remote\.php/);
		$sp_f  = 1 if ($line =~ m/The following.*saml20-sp-remote\.php/);
	}
	close FH;

	return 1 if ($#idp_lines < 0 or $#sp_lines < 0);

	# write saml20-idp-remote.php
	open (FH, ">$idp_meta_conf") or return 1;
	print FH "<?php\n";
	foreach $line (@idp_lines) {
		print FH "$line";
	}
	print FH "?>";
	close FH;

	# write saml20-sp-remote.php
	open (FH, ">$sp_meta_conf") or return 1;
	print FH "<?php\n";
	foreach $line (@sp_lines) {
		print FH "$line";
	}
	print FH "?>";
	close FH;

	return 0;
}

##########################################
# Backup ldap directory to etc of eduroam
#
sub ldap_backup {
	my $back_dest = shift;
	my @lines;
	my $line;

	open (FH, "/usr/sbin/slapcat |") or return 1;
	open (BCK, ">$back_dest") or return 1;
	while(<FH>) {
		$line = $_;
		next if ($line =~ m/^structuralObjectClass:/);
		next if ($line =~ m/^entryUUID:/);
		next if ($line =~ m/^creatorsName:/);
		next if ($line =~ m/^createTimestamp:/);
		next if ($line =~ m/^entryCSN:/);
		next if ($line =~ m/^modifiersName:/);
		next if ($line =~ m/^modifyTimestamp:/);
		print BCK $line;
	}
	close FH;
	close BCK;
	return 0;
}

sub process_ldif {
	# TODO: - detect root dn from ldif - from fist line ($dc_domain_old)
	#       - generate fulldomain ($aai_realm_old)
	#       - replace $dc_domain_old with $dc_domain
	#       - replace @$aai_realm_old with @$aai_realm

}

##########################################
# Output /etc/raddb/clients.conf
#
sub establish_radius_clients {
	my $tmp;
	if (not open(FH, ">/etc/raddb/clients.conf")) {
	        print "ERROR: Unable to insert data to /etc/raddb/clients.conf";
	        exit -1;
	}

	# add upstrem radius servers
	print FH "# ===============================\n";
	print FH "# Upstream radius servers\n";
	print FH "# ===============================\n";
	if ("$eduroam::primary_radius_ip" ne "") {
	        print FH "client $eduroam::primary_radius_ip {\n";
	        print FH "\tsecret    = $eduroam::primary_radius_sec\n";
	        print FH "\tshortname = primary_radius\n";
	        print FH "\tnastype   = other\n";
	        print FH "}\n";
	}
	if ("$eduroam::secondary_radius_ip" ne "") {
	        print FH "client $eduroam::secondary_radius_ip {\n";
	        print FH "\tsecret    = $eduroam::secondary_radius_sec\n";
	        print FH "\tshortname = secondary_radius\n";
	        print FH "\tnastype   = other\n";
	        print FH "}\n";
	}
	print FH "# ===============================\n";
	print FH "# Access Points\n";
	print FH "# ===============================\n";
	foreach $tmp (keys %eduroam::aps) {
	        print FH "client " . $tmp . " {\n";
	        print FH "\tsecret    = " . $eduroam::aps{$tmp}->{"radius_sec"} . "\n";
	        # TODO: we should also store the type and short name (hostname) of AP
	        print FH "\tshortname = apX\n";
	        print FH "\tnastype   = cisco\n";
	        print FH "}\n";
	}
	print FH "# ===============================\n";
	close FH;
}
##########################################
# Output config files from a given templates
# 
sub make_cfgs_from_templates {
	my $template;
	my %template_files = %{$_[0]};
	my $output_file;
	my @output;

	foreach $template (keys %template_files) {
		$output_file = $template_files{$template};
		@output = ();

		print "Processing file $output_file\n";
		open(FH, "<$template");
		while(<FH>) {
			push @output, &process_template_line($_);
		}
		open(FH, ">$output_file");
		print FH @output;
		close(FH);
	}
}

##########################################
# chek if simplesamlphp is installed
# return 1 if it is, 0 if it isn't
sub have_simplesamlphp_rpm {
	my @rpm_out = qx{/bin/rpm -q simplesamlphp};
	return 0 if ((join '', @rpm_out) =~ m/not installed/);
	return 1;
}

##########################################
# gracefully restart service
sub graceful_restart_service {
	my $service = $_[0];
	my $restart = "/sbin/service $service graceful";
	my $tmp;

	print "===================================================\n";
	print "gracefully restarting service $service:\n";
	print "---------------------------------------------------\n";
	print $restart . "\n";
	$tmp = `$restart`;
	print $tmp;

	return;
}
##########################################
# condrestart service
sub condrestart_service {
	my $service = $_[0];
	my $restart = "/sbin/service $service condrestart";
	my $tmp;

	print "===================================================\n";
	print "conditionally restarting service $service:\n";
	print "---------------------------------------------------\n";
	print $restart . "\n";
	$tmp = `$restart`;
	print $tmp;

	return;
}
##########################################
# Activate service
sub activate_service {
	my $service = $_[0];
	my $add = "/sbin/chkconfig --add $service";
	my $on = "/sbin/chkconfig $service on";
	my $restart = "/sbin/service $service restart";
	my $tmp;

	print "===================================================\n";
	print "Activating (restarting) service $service:\n";
	print "---------------------------------------------------\n";
	print $add . "\n";
	$tmp = `$add`;
	print $tmp;
	print $on . "\n";
	$tmp = `$on`;
	print $tmp;
	print $restart . "\n";
	$tmp = `$restart`;
	print $tmp;

	return;
}
##########################################
# stop service
sub deactivate_service {
	my $service = $_[0];
	my $stop = "/sbin/service $service stop";
	my $tmp;

	print "===================================================\n";
	print "Stopping service $service:\n";
	print "---------------------------------------------------\n";
	print $stop . "\n";
	$tmp = `$stop`;
	print $tmp;

	return;
}
##########################################
# Deactivate service
sub stop_service {
	my $service = $_[0];
	my $off = "/sbin/chkconfig $service off";
	my $stop = "/sbin/service $service stop";
	my $tmp;

	print "===================================================\n";
	print "Deactivating (stopping) service $service:\n";
	print "---------------------------------------------------\n";
	print $off . "\n";
	$tmp = `$off`;
	print $tmp;
	print $stop . "\n";
	$tmp = `$stop`;
	print $tmp;

	return;
}
##########################################
sub load_confs {
	my $file;
        my @config_files = @_;

	foreach $file (@config_files) {
		&eduroam::load_conf($file);
	}	
}       

##########################################
sub load_conf {
        # load settings -- the ugly/perl way.
	# prepend path to the config file name
        my $config_file = $eduroam::config_path . "$_[0]";
# 	print "Config file loaded: $config_file\n";

        return "" if (! -f "$config_file");
        open FH, "<$config_file";
        while(<FH>) {
                next if(m/^#/);
                m/(.*)/;
                eval $1;
        }
}       
##########################################
# 1st parameter is hostname
# returns domain name
sub get_domain_from_hostname {
	return undef if &validate_hostname($_[0]);
	$_[0] =~ m/\w*\-*\.(.*)/;
	return $1;
}
##########################################
my $apRadioMACTable = ".1.3.6.1.4.1.9.9.413.1.1.4.1.2.1";
sub get_ap_macs {
	my $ap = $_[0];
	my $community = $_[1];
	my $key;
	my $result;
	my $macs = "";
	my ($session, $error) = Net::SNMP->session(
		-hostname  => $ap,
		-community => $community,
		-translate => [ -octetstring => 0x0 ],
		-version   => "snmpv2c"
	);
                
	return undef if(!defined($session));
	
           
	$result = $session->get_table(
		-baseoid => $apRadioMACTable
	);

	return undef if(!defined($result));
	$session->close;

	# Process returned MAC from AP.
	foreach $key (keys(%{$result})) {
		my $mac = sprintf("%s.%s.%s", unpack("H4H4H4", $result->{$key}));
		$macs .= "$mac ";
        }
	$macs =~ s/\s+$//;
	return $macs;
}
##########################################
# Returns an array of ethX devices on the
# system.
sub get_phy_interfaces {
	my @ifaces;

	open(FD, "</proc/net/dev");
	while(<FD>) {
		if (m/^\s+(eth.*):/) {
			push(@ifaces, "$1");
		}
	}
	return @ifaces;
}
##########################################
# Returns an array of all network devices 
# on the system.
sub get_interfaces {
	my @ifaces;

	open(FD, "</proc/net/dev");
	while(<FD>) {
		if (m/^\s*(.+?):/) {
			next if ("$1" eq "lo"); # everything but loopback
			push(@ifaces, "$1");
		}
	}
	return @ifaces;
}
sub get_interface_cidr {
	my $interface = shift;
	my $cidr="";
	my $up=0;

	open (FH, "/sbin/ip addr show $interface |") or die "error";
	while(<FH>) {
		if (m/$interface.+,UP/) { $up=1; }
		if (m/.*inet\s(.*)\/(\d{1,2}).*/) {
			$cidr = "$1/$2";
			last;
		}
	}
	return $cidr if ($up);
	return "";
}
##########################################
# IpStr2Int
sub IpStr2Int {
	die "assertion failed" if not defined $_[0];
	return undef unless $_[0] =~ m/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/;
	return ($1<<24 | $2<<16 | $3<<8 | $4) & 0xFFFFFFFF;
}
##########################################
# IpInt2Str
sub IpInt2Str {
	return sprintf("%d.%d.%d.%d", $_[0]>>24&0xFF, $_[0]>>16&0xFF, $_[0]>>8&0xFF, $_[0]&0xFF);
}
##########################################
# Returns network addx and netmask from cidr
sub IpCidr2NetAndMask {
	my @res;
	my $ip;
	my $nmask;
	@res = &IpCidr2IpNetmask($_[0]);
	return undef unless @res;
	$ip = &IpStr2Int($res[0]);
	$nmask = &IpStr2Int($res[1]);
	$ip = $ip & $nmask; # make network address out of it
	$ip = &IpInt2Str($ip);
	return ($ip, $res[1]);
}
##########################################
# Returns netmask from IP/CIDR
sub IpCidr2IpNetmask {
	my $cidr;
	my $ip;
        my $str_net = $_[0];
	my $bit_mask;
	my $nmask;
	return undef unless ($str_net =~ m/^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\/(\d{1,2})$/);
	$ip = $1;
	$cidr = 0+$2;
	return undef unless ($cidr>=0 && $cidr<=32);
	$bit_mask  = (0xFFFFFFFF<<(32-$cidr));
	$bit_mask &=  0xFFFFFFFF;
	$nmask = &IpInt2Str($bit_mask);
	return ($ip, $nmask);
}
##########################################
# Checks if address is within the given subnet
# 1st IP
# 2nd CIDR
# returns 0 for Yes and !0 for No
sub IpCmp {
	my $srcip = $_[0];
	my $cidr = $_[1];
	my $dstip;
	my $dstmask;
	return 1 if not defined $srcip;
	return 1 if not defined $cidr;
	return 1 if length $srcip  <= 0;
	return 1 if length $cidr   <= 0;
	($dstip, $dstmask) = &IpCidr2IpNetmask($cidr);
	return (&IpStr2Int($dstip) | ~&IpStr2Int($dstmask)) - (&IpStr2Int($srcip) | ~&IpStr2Int($dstmask));
}
##########################################
# Return min and max usable address
sub IpCidr2MinMaxIP {
	my $ipmin;
	my $ipmax;
	my $cidr;
	my $ip;
	my $bit_mask;
        my $str_net = $_[0];

	return undef unless $str_net =~ m/^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\/(\d{1,2})$/;
	$ip = &IpStr2Int($1);
	return undef if (!defined $ip);
	$cidr = 0+$2;
	return undef unless ($cidr>=0 && $cidr<=32);
	$bit_mask  = (0xFFFFFFFF<<(32-$cidr));
	$bit_mask &=  0xFFFFFFFF;
	$ipmin     =  ($ip &  $bit_mask) +1;
	$ipmax     =  ($ip | ~$bit_mask) -1;
	$ipmin     = &IpInt2Str($ipmin);
	$ipmax     = &IpInt2Str($ipmax);
	return ($ipmin, $ipmax);
}

##########################################
# Executes a command or prints out a HTML error message.
sub system_or_htmldie {
	my $message;
	my $error = $_[1];
	my $tmp;
	my $i = 0;
	
	$error = "Failed to execute $_[0]" if !defined $_[1];
# 	print "<br>\n";
	STDOUT->autoflush(1);
	&html_die($error) if (not open (FH, "$_[0] 2>&1 |"));
	FH->autoflush(1);
	print "<pre>";
	while(0 != sysread FH, $tmp, 1) {
		if("$tmp" eq "\n") {
			$i = 0;
			print "\n";
			next;
		}
		print $tmp;
		$i++;
		print "\n" if (0==$i%100);
	}
	close FH;
	print "</pre>\n";
}
#sub system_or_htmldie {
#	my $message;
#	my $error = $_[1];
#                        
#	$error = "Failed to execute $_[0]" if !defined $_[1];
#                
#	$message = `$_[0] 2>&1`;
#	if ($?) {
#		print "Content-Type: text/html\n\n";
#		print "ERROR: $error<br>\n";
#		if (defined $message and length $message > 0) {
#			$message =~ s/\n/<br>/g;
#			print "Detailed error report:<br>$message";
#		}
#		exit -1; 
#	}       
#	return $message;
#}               
##########################################
# Report the error passed as the first parameter and die.
# Second (optional) parameter is a web form to return to.
# Third (optional) parameter is the end HTML.
# ----------------------------------------
# Content-Type and maybe some header must
# have been already printed out.
sub html_die {
	print "Critical ERROR: $_[0]<BR>\n";
	print "Check web server log files.\n";
	if(defined $_[1]) {
		print "<FORM METHOD=POST ACTION=\"$_[1]\">";
		print "<INPUT TYPE=\"submit\" NAME=Confirm VALUE=Confirm><br>";
		print "</FORM>";
	}
	print $_[2] if (defined $_[2]);
	exit -1;
} 
##########################################
sub html_perror {
	print "Content-Type: text/html\n\n";
	print "$_[0]<BR>\n";
	print "<FORM METHOD=POST ACTION=\"$_[1]\">";
	print "<INPUT TYPE=\"submit\" NAME=Confirm VALUE=Confirm><br>";
	print "</FORM>";
	exit 0;
}
##########################################
sub html_perror2 {
	print "$_[0]<BR>\n";
	print "<FORM METHOD=POST ACTION=\"$_[1]\">";
	print "<INPUT TYPE=\"submit\" NAME=Confirm VALUE=Confirm><br>";
	print "</FORM>";
	exit 0;
}
##########################################

return 1;

END {}
