#!/usr/bin/perl 
#
# Server's tasks:
# - tail dhcp log and log complement logs in MySQL with IP information
# - check active/stale user connections on APs and close old accounting records (stale connections)
# - tail radius log file and insert incorrect login attempts in MySQL table
#
# We need the following to run:
# MySQL user 'apd' with password $eduroam::apd_pw
# MySQL user 'dhcpd' with password $eduroam::dhcpd_pw
# MySQL user 'denied' with password $eduroam::denied_pw
# APs (IP, snmp_password, root user password)
#
# VERSION: 20081230

use strict;
use File::Tail;
use Time::Local;
use Sys::Syslog;
use POSIX;
use DBI;
use Net::SNMP;
use Fcntl ':flock';

############################################
# Configuration                            # 
############################################
my $dsn = 'DBI:mysql:radius:localhost';
my $sql_user_name_dhcp = 'dhcpd';
my $sql_user_name_apd = 'apd';
my $sql_user_name_denied = 'denied';
my $logfile = '/var/log/dhcpd.log';
my $radius_log = '/var/log/radius/radius.log';
my $pid_file = '/var/run/eduroam_monitor.pid';

# Which facility should messages be logged to
my $facility = 'local6'; # write log to the facility X.
my $check_conn = 10*60; # Check existing connection every X seconds: 10 minutes.
my $online_time_check = 60*60*24; # 24 hrs
#my $online_time_check = 5; # 24 hrs

# Load "libraries"
require '/usr/lib/eduroam/eduroam.pm';

# obsolete... we use MySQL database for that.
#my $config_file1 = "/var/eduroam/etc/ap.conf";
my $config_file2 = "sys_db";
my $config_file3 = "net.conf";

# ==========================================
# and constants/variables
my $oidSysDescr = ".1.3.6.1.2.1.1.1.0";
my $oidCiscoStationTable = ".1.3.6.1.4.1.9.9.273.1.2.1.1.14";
my $oidLancomStationTable = ".1.3.6.1.4.1.2356.600.2.54.1.3.32.1";
#my $apClientTable = ".1.3.6.1.4.1.9.9.273.1.2.1.1.14";
my $apRadioMACTable = ".1.3.6.1.4.1.9.9.413.1.1.4.1.2.1";
my %months = ('Jan', 0, 'Feb', 1, 'Mar', 2, 'Apr', 3, 'May',  4, 'Jun',  5,
	      'Jul', 6, 'Aug', 7, 'Sep', 8, 'Oct', 9, 'Nov', 10, 'Dec', 11);
my $pid_dhcp = 0;
my $pid_ap = 0;
my $pid_denied = 0;
my $logid = "M:";
my $file;
my $line;
my $rc;
my $dbh;
my $sql_dhcp;
my $sql_denied;
my $sql_get;
my $sql_close;
my %ap_mac_ip = ();

# ==========================================
sub signal_handler {
    syslog('err', "$logid Received signal $_[0], TERMINATING!");
    unlink($pid_file);
    exit(-1);
}
# ==========================================
# cisco_mac - convert to cisco MAC format
sub cisco_mac {
	my $in_mac = $_[0];
	if ($in_mac =~ m/([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})/) {
		return sprintf("%02x%02x.%02x%02x.%02x%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);;
	}
	return  $in_mac;
}
# ==========================================
# lancom_mac - convert to lancom MAC format
sub lancom_mac {
	my $in_mac = $_[0];
	if ($in_mac =~ m/([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})\:([0-9,A-F,a-f]{2})/) {
		return sprintf("%02x-%02x-%02x-%02x-%02x-%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);;
	}
	return  $in_mac;
}
# ==========================================
# canonical_mac - convert to ordinary MAC format
# input:   xx-xx-xx-xx-xx-xx   -or-
#          xxxx-xxxx-xxxx      -or-
#	   x:x:x:x:x:x         -or- ...
# output:  xx:xx:xx:xx:xx:xx
# Return what you got if string is unrecognisable

sub canonical_mac {
	my $in_mac = $_[0];
	
	# Ordinary MAC (can be in slim or fat format, dot colon or dash seperated)
	if ($in_mac =~ m/([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})/) {
		return sprintf("%02x:%02x:%02x:%02x:%02x:%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);;
	}

	# Cisco MAC
	if ($in_mac =~ m/([0-9,A-F,a-f]{2})([0-9,A-F,a-f]{2})\.([0-9,A-F,a-f]{2})([0-9,A-F,a-f]{2})\.([0-9,A-F,a-f]{2})([0-9,A-F,a-f]{2})/) {
		return sprintf("%02x:%02x:%02x:%02x:%02x:%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);;
	}

	return $in_mac;
}
# ==========================================
# convert_cisco_mac - convert MAC from cisco format
#
sub convert_cisco_mac {
        if ($_[0] =~ m/([0-9,A-F,a-f]{2})([0-9,A-F,a-f]{2})\.([0-9,A-F,a-f]{2})([0-9,A-F,a-f]{2})\.([0-9,A-F,a-f]{2})([0-9,A-F,a-f]{2})/) {
                return sprintf("%02x:%02x:%02x:%02x:%02x:%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);
        }
        return $_[0];
}
# ==========================================
# slim_mac - convert MAC to slim format
#
sub slim_mac {
        my $mac = &convert_cisco_mac($_[0]);
        unless ($mac =~ m/^([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})$/)
		{ return undef };
        return sprintf("%x:%x:%x:%x:%x:%x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);
}
# ==========================================
# fat_mac - convert MAC to fat format
#
sub fat_mac {
        my $mac = &convert_cisco_mac($_[0]);
        unless ($mac =~ m/^([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})[.:-]([0-9,A-F,a-f]{1,2})$/)
		{ return undef };
        return sprintf("%02x:%02x:%02x:%02x:%02x:%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);
}

# ==========================================
# sql_get_snmp - fetches snmp secret for a
# IP address of the Access Point
sub sql_get_snmp {
        my $tmp = "";
        my $result;
	my $dbh;        # TODO: use existing connection to the database!!!
	my $num;
	my $sql_query;

        $dbh = DBI->connect($dsn, $sql_user_name_apd, $eduroam::apd_pw);
	if (not defined $dbh) {
		syslog('err', "$logid ERROR in sql_users_get(), unable to connect to database");
		return undef;
	}

        $sql_query = $dbh->prepare(q{
	SELECT `snmp secret`
        FROM access_points WHERE
		`IP address` = ?
	LIMIT 1}
	);
        
        $sql_query->execute("$_[0]") || die $dbh->errstr;
	syslog('err', "$logid SQL query for " . $_[0] . " returned no results") if ($sql_query->rows <= 0);
        while ($result = $sql_query->fetchrow_arrayref) {
                # The data returned is: `snmp secret`
		#syslog('debug', "$logid SQL AP " . $_[0] . " snmp $result->[0]");
		$tmp = $result->[0];
        }

        $dbh->disconnect;
        return $tmp;
}
# ==========================================
# sql_users_get - fetches all active session
# records from the SQL database and returns
# them in a hash table %users.
# 
sub sql_users_get {
        my $tmp;
        my $key;
        my $data;
        my $result;
        my %users;      # Key is a MAC, data is array[IP,IP, ...]
	my $dbh;        # TODO: use existing connection to the database!!!

        $dbh = DBI->connect($dsn, $sql_user_name_apd, $eduroam::apd_pw);
	if (not defined $dbh) {
		syslog('err', "$logid ERROR in sql_users_get(), unable to connect to database");
		return undef;
	}

        $sql_get = $dbh->prepare(q{SELECT 
                `Calling-Station-Id`,
                `Client-IP-Address`
        FROM ACCOUNTING WHERE
                `Timestamp Stop` = FROM_UNIXTIME(0) 
        AND     `Client-IP-Address` != ''
        ORDER BY
                `Timestamp Start` DESC
        });
        
        $sql_get->execute() || die $dbh->errstr;
        while ($result = $sql_get->fetchrow_arrayref) {
                # The data returned is: Calling-Station-Id, Client-IP-Address
		syslog('debug', "$logid SQL active session: $result->[0] ($result->[1])");
                $tmp = $users{$result->[0]};
                push(@$tmp, $result->[1]);
                $users{$result->[0]} = $tmp;
        }

        $dbh->disconnect;
        return %users;
}
# ==========================================
# Create daemon and setup syslog logging
sub daemonize {
    my $pid = 1;
    my $rc = 0;
    openlog('eduroam_monitor', 'ndelay,pid', $facility)
        or die "Can't open syslog: $!";
    chdir '/'                  or die "Can't chdir to /: $!";
    umask 0;
    open STDIN, '/dev/null'    or die "Can't read /dev/null: $!";
    setsid                     or die "Can't start a new session: $!";
    open STDOUT, '>>/dev/null' or die "Can't write to /dev/null: $!";
    open STDERR, '>>/dev/null' or die "Can't write to /dev/null: $!";
    $SIG{INT}='signal_handler';
    $SIG{TERM}='signal_handler';
    $SIG{KILL}='signal_handler';
    $SIG{__DIE__}='signal_handler';
    $pid = fork;
    defined($pid) or syslog('err', "$logid Can't fork: $!");
    if ($pid) {
        open PIDFILE, ">$pid_file" or die "can't open $pid_file: $!\n";
        print PIDFILE $pid;
        close PIDFILE;
        exit 0;
    }
}
# ==========================================
sub connect_to_mysql_dhcp {
	$dbh = DBI->connect($dsn, $sql_user_name_dhcp, $eduroam::dhcpd_pw);
	if(!defined($dbh)) {
		syslog('err', "$logid ERROR: connect to MySQL database failed!\n");
		exit(-1);
	}
	
        $sql_dhcp = $dbh->prepare(q{UPDATE ACCOUNTING SET 
                `Client-IP-Address` = ?,
                `Timestamp Dhcp` = NOW()
        WHERE
                (`Calling-Station-Id` = ? OR `Calling-Station-Id` = ?) AND
                `Timestamp Start` <= FROM_UNIXTIME(?) AND
                `Timestamp Stop` = FROM_UNIXTIME(0) 
        });
}
# ==========================================
sub connect_to_mysql_ap {
	$dbh = DBI->connect($dsn, $sql_user_name_apd, $eduroam::apd_pw);
	if(!defined($dbh)) {
		syslog('err', "$logid ERROR: connect to MySQL database failed!\n");
		exit(-1);
	}

        $sql_get = $dbh->prepare(q{SELECT 
                `Calling-Station-Id`,
                `User-Name`,
                UNIX_TIMESTAMP(`Timestamp Start`),
                `Client-IP-Address`,
                `NAS-IP-Address`
        FROM ACCOUNTING WHERE
                `Timestamp Stop` = FROM_UNIXTIME(0) 
        ORDER BY
                `Timestamp Start` DESC
        });

        $sql_close = $dbh->prepare(q{UPDATE ACCOUNTING SET
                `Timestamp Stop`         = NOW(),
                `Acct-Terminate-Cause`   = 'Script-Cleanup'
        WHERE   `Calling-Station-Id` = ?
        AND     `User-Name` = ?
        AND     `Client-IP-Address` = ?
        AND     `NAS-IP-Address` = ?
        AND     `Timestamp Stop` = FROM_UNIXTIME(0) 
        AND     `Timestamp Start` = FROM_UNIXTIME(?)
        });
}
# ==========================================
sub connect_to_mysql_denied {
	$dbh = DBI->connect($dsn, $sql_user_name_denied, $eduroam::denied_pw);
	if(!defined($dbh)) {
		syslog('err', "$logid ERROR: connect to MySQL database failed!\n");
		exit(-1);
	}
	
	$sql_denied = $dbh->prepare(q{INSERT INTO denied 
		(`User-Name`, `Calling-Station-Id`, `NAS-Shortname`, `NAS-Port`, `Timestamp`, `Cause`)
		VALUES
		(?, ?, ?, ?, NOW(), ?)
	});
}
# ==========================================
sub get_ap_type {
	my $session = $_[0];
	my $result;
	
#      $session->hostname, $result->{$sysUpTime} 

	$result = $session->get_request(-varbindlist => [$oidSysDescr]);
	if (defined($result)) {
                #syslog('debug',"$logid sysDescr returned: %s\n", $result->{$oidSysDescr});
		return "lancom" if ($result->{$oidSysDescr} =~ m/LANCOM L-54/);
		return "cisco" if ($result->{$oidSysDescr} =~ m/Cisco IOS Software/);
	} else {
                syslog('err',"$logid ERROR in SNMP get_request for Access point system description: %s\n", $session->error);
	}
	return "unknown";
}
# ==========================================
sub check_client_cisco {
	my $session = $_[0];
        my $mac = $_[1];
        my $mac_oid;
        my $result;
	my $key;

        syslog('debug',"$logid Cisco check_client: %s\n", $mac);
        $mac =~ m/(..)\:(..)\:(..)\:(..)\:(..)\:(..)/;
        $mac_oid = sprintf("%d.%d.%d.%d.%d.%d", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);

	$result = $session->get_table(
                -baseoid => $oidCiscoStationTable 
        );
        if(!defined($result)) {
                if(!$session->error_index) {
                        return 0; # an empty table "error", that's OK.
                }
                # Complain and that's it. Return as if user is present.
                syslog('err',"$logid ERROR in SNMP get_table: %s\n", $session->error);
                return 1;
        }

        foreach $key (keys(%{$result})) {
                if ($key =~ m/$mac_oid$/) {
                	syslog('debug', "$logid Cisco check $mac, state = %d \n", $result->{$key});
                        return $result->{$key} == 3;
                }
        }
        return 0;
}
# ==========================================
sub check_client_lancom {
	my $session = $_[0];
        my $mac = $_[1];
	my $tmp_mac;
	my $hexmac;
        my $result;
	my $key;

        syslog('debug',"$logid Lancom check_client: %s\n", $mac);
        $mac =~ m/(..)\:(..)\:(..)\:(..)\:(..)\:(..)/;
	$hexmac = sprintf("0x%02x%02x%02x%02x%02x%02x", hex $1, hex $2, hex $3, hex $4, hex $5, hex $6);
	$result = $session->get_table(
                -baseoid => $oidLancomStationTable
        );
        if(!defined($result)) {
                if(!$session->error_index) {
                        return 0; # an empty table "error", that's OK.
                }
                # Complain and that's it. Return as if user is present.
                syslog('err',"$logid ERROR in SNMP get_table: %s\n", $session->error);
                return 1;
        }

	# Lancom has MAC address in variable .X.4.1 where X is association ID.
        foreach $key (keys(%{$result})) {
                #syslog('debug',"$logid Lancom check: %s val: %s\n", $key, $result->{$key});
		next unless ($key =~ m/^(.*)\.4\.1$/); # only interested in MAC (.4.1)
		$tmp_mac = sprintf('0x%s', unpack('H*', $result->{"$1.4.1"}));
		next unless ($tmp_mac eq $hexmac); # Is this the client/MAC we are seaching for ?
		return $result->{"$1.10.1"} == 3;
        }

	return 0;
}
# ==========================================
# Check if client with a given MAC is online 
# 1st parameter: Client MAC
# 2nd parameter: AP IP
sub check_client {
        my $mac;
        my $ap_ip = $_[1];
	my $ap_community;
	my $ap_type;
        my $key;
        my $tmp;

	$mac = &canonical_mac($_[0]); # make canonical format
	#$ap_community = $eduroam::aps{$ap_ip}->{'snmp'}; # Get AP SNMP password
	$ap_community = &sql_get_snmp($ap_ip); # Get AP SNMP password
 
        my ($session, $error) = Net::SNMP->session(
                -hostname  => $ap_ip,
                -community => $ap_community,
                -translate => [ -octetstring => 0x0 ],
                -version   => "snmpv2c"
        );

        if(!defined($session)) {
                syslog('err', "$logid ERROR in SNMP session establish: %s\n", $error);
                return 1;
        }

	# Query to see which ap it is ?
	$ap_type = get_ap_type($session);

	if ($ap_type eq "cisco") {
		$tmp = &check_client_cisco($session, $mac);
		$session->close;
		return $tmp;
	}
	
	if ($ap_type eq "lancom") {
		$tmp = &check_client_lancom($session, $mac);
		$session->close;
		return $tmp;
	}

        $session->close;
	syslog('err',"$logid ERROR. Access Point $ap_ip is not supported!\n", $session->error);
	return 1; # return as if user is present
}
# ==========================================
# Monitor dhcp log file and insert IP address
sub monitor_dhcp {
	my $rc;

	&connect_to_mysql_dhcp;
	
	$file=File::Tail->new(name=>$logfile, maxinterval=>1, interval=>1);
	syslog('notice',"$logid DHCP monitor running, PID: $$\n");
	
	while( 1 ) {
		$line = $file->read;
		if(!defined($line)) {
			syslog('err',"$logid Unable to get a line from DHCP log! TERMINATING!");
			exit(-1);
		}
		$rc = $dbh->ping;
		if(1 != $rc) {
			syslog('warning',"$logid Lost connection to MySQL database! Restarting the connection!");
			$dbh->disconnect;		
			&connect_to_mysql_dhcp;
			$rc = $dbh->ping;
			if(1 != $rc) {
				syslog('err',"$logid Again lost connection to MySQL database! TERMINATING!");
				exit(-1);
			}
		}
	
		if ( $line =~ m/^(\w+)\s+(\d+)\s+(\d+)\:(\d+)\:(\d+).*dhcpd\: DHCPACK.on.(\S+).to.(\S+).*via.(\S+)/ ) {
			my $month = $1;
			my $day = $2;
			my $h = $3;
			my $m = $4;
			my $s = $5;
			my $ip = $6;
			my $mac = &canonical_mac($7);
			my $iface = $8;
			my $year = (localtime)[5] + 1900;
			my $cisco_mac;
			my $lancom_mac;

			# server as eduroam gateway is no longer supported
			# TODO: find some other way to seperate eduroam form other users
# 			# Check if this is a request for a "managed" VLAN/interface. 
# 			next unless ($iface =~ m/^\w+\.$eduroam::user_vlan$/);

			$month = $months{$month};
			# check year change!
			if( 11 == $month && 31 == $day ) {
				my $cur_day = (localtime)[3];
				# If it's january the 1st, we missed a year change!
				if( 1 == $cur_day) {
					$year--;
				}
			}
			# change MAC to supported formats... 
			$cisco_mac = &cisco_mac($mac);
			$lancom_mac = &lancom_mac($mac);
			my $timestamp = timelocal($s, $m, $h, $day, $month, $year);
			$rc = $sql_dhcp->execute("$ip", "$cisco_mac","$lancom_mac", $timestamp ) || die $dbh->errstr;
			if ($rc < 1) {
				syslog('err', "$logid FATAL ERROR inserting DHCP assigned address: $year-%02d-%02d, %02d:%02d:%02d, IP: $ip, MAC: $mac, timestamp: $timestamp", $month + 1, $day, $h, $m, $s);
				exit -1;
			}
			syslog('info', "$logid Inserted DHCP assigned address: $year-%02d-%02d, %02d:%02d:%02d, IP: $ip, MAC: $mac, timestamp: $timestamp", $month + 1, $day, $h, $m, $s);
			#$dbh->commit || die $dbh->errstr;
		}
	}
	$dbh->disconnect;
	exit(0);
}
# ==========================================
# Monitor radius log file and insert failed login attempts
sub monitor_denied {
	my $rc;

	&connect_to_mysql_denied;

	$file=File::Tail->new(name=>$radius_log, maxinterval=>1, interval=>1);
	syslog('notice',"$logid denied login monitor running, PID: $$\n");

	my %buffer = ();

	while( 1 ) {
		$line = $file->read;
		if(!defined($line)) {
			syslog('err',"$logid Unable to get a line from radius log! TERMINATING!");
			exit(-1);
		}
		$rc = $dbh->ping;
		if(1 != $rc) {
		syslog('warning',"$logid Lost connection to MySQL database! Restarting the connection!");
			$dbh->disconnect;		
			&connect_to_mysql_denied;
			$rc = $dbh->ping;
			if(1 != $rc) {
				syslog('err',"$logid Again lost connection to MySQL database! TERMINATING!");
				exit(-1);
			}
		}

		# Login incorrect (Home Server says so): [anon-User-Name] (from client NAS-Shortname port PORT cli MAC)
		if ($line =~ m/Login incorrect \(Home Server says so\)\: \[(.+)\] \(from client (\S+) port (\d+)( cli \S+)?\)/) {
			my $user_name = $1;
			my $nas_shortname = $2;
			my $nas_port = $3;
			my $mac = $4;
			$mac =~ s/ cli //;
			$mac = &canonical_mac($mac);
			my $cisco_mac = &cisco_mac($mac);
			my $cause = "Rejected by user's home server";

			%buffer = ();

			$rc = $sql_denied->execute("$user_name", "$cisco_mac", "$nas_shortname", "$nas_port", "$cause") || die $dbh->errstr;
			if ($rc < 1) {
				syslog('err', "$logid FATAL ERROR inserting denied login: User-Name: $user_name, MAC: $mac");
				exit -1;
			}
			syslog('info', "$logid Inserted denied login: User-Name: $user_name, MAC: $mac");

			next;
		}

# 		if ( $line =~ m/Attribute \"Password\" is required for authentication/ ) {
# 			$buffer{'cause'} = "Empty password";
# 			next;
# 		}

		#### Inner ####

		# FreeRadius 1.x version
		# Login incorrect (Cause): [User-Name] (from client loopback port 0)
		if ($line =~ m/Login incorrect \((.+)\)\: \[(.+)\] \(.+loopback/) {
			if (exists $buffer{'cause'} ) {
				$buffer{'cause'} = $buffer{'cause'}." / $1"; 
			} else { $buffer{'cause'} = $1; }
			$buffer{'user_name'} = $2;
			next;
		}

		# FreeRadius 2.x version
		# Login incorrect (Cause): [User-Name] (from client NAS-Shortname port PORT cli MAC via TLS tunnel)
		if ($line =~ m/Login incorrect \((.+)\)\: \[(.+)\] \(from client (\S+) port (\d+)( cli \S+)? via TLS tunnel\)/) {
			if (exists $buffer{'cause'} ) {
				$buffer{'cause'} = $buffer{'cause'}." / $1"; 
			} else { $buffer{'cause'} = $1; }
			$buffer{'user_name'} = $2;
			next;
		}

		# FreeRadius 1.x version
		# Login incorrect: [User-Name] (from client loopback port 0)
		if ($line =~ m/Login incorrect\: \[(.+)\] \(.+loopback/) {
			$buffer{'user_name'} = $1;
			next;
		}

		# FreeRadius 2.x version
		# Login incorrect: [User-Name] (from client NAS-Shortname port PORT cli MAC via TLS tunnel)
		if ($line =~ m/Login incorrect: \[(.+)\] \(from client (\S+) port (\d+)( cli \S+)? via TLS tunnel\)/) {
			$buffer{'user_name'} = $1;
			next;
		}


		#### Outer ####

		# FreeRadius 2.x version
		# TLS Alert read:warning:bad certificate
		if ($line =~ m/TLS Alert.*bad certificate/) {
			if (exists $buffer{'cause'} ) {
				$buffer{'cause'} = $buffer{'cause'}." / Bad CA certificate on client"; 
			} else { $buffer{'cause'} = "Bad CA certificate on client"; }
			next;
		}

		# FreeRadius 1.x & 2.x version
		# Login incorrect: [anon-User-Name] (from client NAS-Shortname port PORT cli MAC)
		if ( $line =~ m/Login incorrect\: \[(.+)\] \(from client (\S+) port (\d+)( cli \S+)?\)/ ) {
			my $user_name = $1;
			my $nas_shortname = $2;
			my $nas_port = $3;
			my $mac = $4;
			$mac =~ s/ cli //;
			$mac = &canonical_mac($mac);
			my $cisco_mac = &cisco_mac($mac);
			my $cause;

			if (exists $buffer{'cause'} ) {
				$cause = $buffer{'cause'};
			}
# 			if ($user_name !~ m/^(anonymous)?\@.+$/) {
# 				if ($cause ne "") { $cause = " / $cause"; }
# 				$cause = "Invalid outer identity ($user_name)".$cause;
# 			}
			if (exists $buffer{'user_name'} ) {
				$user_name = $buffer{'user_name'};
			}
			%buffer = ();

			$rc = $sql_denied->execute("$user_name", "$cisco_mac", "$nas_shortname", "$nas_port", "$cause") || die $dbh->errstr;
			if ($rc < 1) {
				syslog('err', "$logid FATAL ERROR inserting denied login: User-Name: $user_name, MAC: $mac");
				exit -1;
			}
			syslog('info', "$logid Inserted denied login: User-Name: $user_name, MAC: $mac");
		}
	}
	$dbh->disconnect;
	exit(0);
}
# ==========================================
# Check connections on every interval
sub monitor_connections {
	my $rc;
	my $tmp;
	my $key;
	my $data;
	my $result;
	my %users;	# Key is a MAC, data is array[array[User, Start, IP, AP MAC]]

	syslog('notice',"$logid Access Point monitor running, PID: $$\n");
	&connect_to_mysql_ap;

	while (1) {
		%users = {};
		$rc = $dbh->ping;
		if(1 != $rc) {
			syslog('warning',"$logid Lost connection to MySQL database! Restarting the connection!");
			$dbh->disconnect;		
			&connect_to_mysql_ap;
			$rc = $dbh->ping;
			if(1 != $rc) {
				syslog('err',"$logid Again lost connection to MySQL database! TERMINATING!");
				exit(-1);
			}
		}

		$sql_get->execute() || die $dbh->errstr;
		while ($result = $sql_get->fetchrow_arrayref) {
        		# The data returned is: Calling-Station-Id, User-Name, Timestamp Start, Client-IP-Address, NAS-IP-Address 
	        	syslog('debug', "$logid User $result->[1], mac: $result->[0] on $result->[4], Start: $result->[2]\n");
	        	$tmp = $users{$result->[0]}; # we want to append a record to an already existing key.
	        	push(@$tmp, [$result->[1], $result->[2], $result->[3], $result->[4]]);
	        	$users{$result->[0]} = $tmp;
		}               

		foreach $key (keys(%users)) {
        		my $i = 0;
			my $rows = 0;
        		$tmp = $users{$key};
	        	for $data (@$tmp) {
        	        	syslog('debug', "$logid MAC($i): $key, USER: $data->[0], Start: $data->[1], IP: $data->[2], AP IP: $data->[3]\n");
	        	        if(!$i and $data->[1] + $online_time_check < time()) {
					syslog('debug',"$logid Record IS OLD, checking user\n");
					if(!check_client($key, $data->[3])) {
                                		syslog('debug', "$logid User not present, closing the accounting record\n");
	                                	$rows = $sql_close->execute($key, $data->[0], $data->[2], $data->[3], $data->[1]) || die $dbh->errstr;
						die "ERROR: SQL update request failed!" if (0 == $rows);
	        	        		syslog('info', "$logid Accounting closed. MAC: $key, USER: $data->[0], Start: $data->[1], IP: $data->[2], AP IP: $data->[3]\n");
        	                	} else {
	                                	syslog('debug', "$logid User is still connected to the AccessPoint\n");
					}
                		}
                		if ($i) {
                        		syslog('debug', "$logid Record is a duplicate, checking user\n");
                        		if(!check_client($key, $data->[3])) {
	                                	syslog('debug', "$logid User not present, closing the accounting record\n");
        	                        	$rows = $sql_close->execute($key, $data->[0], $data->[2], $data->[3], $data->[1]) || die $dbh->errstr;
						die "ERROR: SQL update request failed!" if (0 == $rows);
	        	        		syslog('info', "$logid Accounting closed. MAC: $key, USER: $data->[0], Start: $data->[1], IP: $data->[2], AP IP: $data->[3]\n");
                	        	} else {
	                                	syslog('debug', "$logid User is connected to the AccessPoint\n");
					}
                		}
			$i++;
        		}
		} # Checking users
		sleep($check_conn);
	}

	$dbh->disconnect;
	exit(0);
}

# ==========================================

# Load settings
#&eduroam::load_conf($config_file1);
&eduroam::load_conf($config_file2);
&eduroam::load_conf($config_file3);

# Fork as a daemon and set-up logging to syslog.
&daemonize;

while(1) {
	# If child doesn't exist, create it.
	if(!$pid_dhcp) {
		$pid_dhcp = fork;
		defined($pid_dhcp) or syslog('err', "$logid FATAL: Can't fork: $!");
		if(!$pid_dhcp) {
			$logid = "D:";
			monitor_dhcp();	
		}
	}
	if(!$pid_ap) {
		$pid_ap = fork;
		defined($pid_ap) or syslog('err', "$logid FATAL: Can't fork: $!");
		if(!$pid_ap) {
			$logid = "A:";
			monitor_connections();	
		}
	}
	if(!$pid_denied) {
		$pid_denied = fork;
		defined($pid_denied) or syslog('err', "$logid FATAL: Can't fork: $!");
		if(!$pid_denied) {
			$logid = "R:";
			monitor_denied();	
		}
	}

	$rc = wait();
        if($rc <= 0) {
		syslog('err', "$logid Error $rc, waiting for child");
		exit -1;
	}
	if ($rc == $pid_dhcp) {
		syslog('warning', "$logid DHCP child $pid_dhcp died, restarting it!");
		$pid_dhcp = 0;
		sleep(5); # So the process doesn't go wild && crazy
	}
	if ($rc == $pid_ap) {
		syslog('warning', "$logid AP child $pid_ap died, restarting it!");
		$pid_ap = 0;
		sleep(5); # So the process doesn't go wild && crazy
	}
	if ($rc == $pid_denied) {
		syslog('warning', "$logid denied child $pid_denied died, restarting it!");
		$pid_denied = 0;
		sleep(5); # So the process doesn't go wild && crazy
	}
}

exit 0;
